#ifndef UU_OBJECTS_LOOPMODE_H_
#define UU_OBJECTS_LOOPMODE_H_

namespace uu {
namespace net {

/**
 * Specifies whether loops are allowed or not
 */
enum class LoopMode
{
    ALLOWED,
    DISALLOWED
};

}
}

#endif
